# -*- coding: utf-8 -*-
from .autosig import autosig, Signature, signature, param
"""Top-level package for autosig."""

__all__ = ["autosig", "Signature", "signature", "param"]
__author__ = """Antonio Piccolboni"""
__email__ = 'autosig@piccolboni.info'
__version__ = '0.2.2'



"""Tests for autosig."""
from hypothesis import given
from hypothesis.strategies import builds, text, dictionaries, just
from string import ascii_letters
from attr import make_class, asdict, attrib
from functools import partial
from autosig import Signature, autosig

# hypothesis strategy for identifiers
# min_size is 10 to avoid hitting reserved words by mistake
identifiers = partial(text, alphabet=ascii_letters, min_size=10)


def signatures():
    """Short summary.

    Returns
    -------
    type
        Description of returned object.

    """
    return builds(
        make_class,
        name=identifiers(),
        attrs=dictionaries(
            keys=identifiers(), values=just(attrib(default=True)), min_size=3),
        bases=just((Signature, )))


@given(sig=signatures())
def test_sig_checks(sig):
    """Autosig-decorated functions accept a compatible set of arguments."""

    @autosig(Sig=sig)
    def f(*args, **kwargs):
        pass

    f(**asdict(sig()))


@given(sig1=signatures(), sig2=signatures())
def test_sig_doesnt_check(sig1, sig2):
    """Autosig-decorated functions fail on an incompatible set of arguments."""

    @autosig(Sig=sig1)
    def f(*args, **kwargs):
        pass

    try:
        f(**asdict(sig2()))
    except Exception:
        return
    raise Exception



"""Implementation of autosig."""
from attr import attrs as signature
from attr import attrib as param
from attr import asdict
from functools import wraps
import inspect

__all__ = ["Signature", "signature", "autosig", "param"]


@signature
class Signature:
    """Base class for signatures."""

    def validate(self):
        """Short summary.

        Returns
        -------
        type
            Description of returned object.

        """
        return True

    def default(self):
        pass


def autosig(Sig):
    """Decorate  functions to attach signatures.

    Parameters
    ----------
    Sig : A subclass of Signature
        A class with one member per parameter, initialized with a call to param

    Returns
    -------
    function
        The decorated function, will intialize, convert, and
        validate its arguments.

    """

    def decorator(f):
        f_params = inspect.signature(f).parameters
        Sig_params = inspect.signature(Sig).parameters
        assert f_params == Sig_params, "\n".join([
            "Mismatched signatures:",
            str(f),
            str(f_params),
            str(Sig),
            str(Sig_params)
        ])

        @wraps(f)
        def wrapped(*args, **kwargs):
            params = Sig(**inspect.getcallargs(f, *args, **kwargs))
            params.validate()
            return f(**asdict(params))

        return wrapped

    return decorator
